package lightcontainer.domains.client;

import lightcontainer.interfaces.ClientHandlerFFE;
import lightcontainer.interfaces.ProtocolRepository;
import lightcontainer.protocol.ProtocolReader;
import lightcontainer.protocol.ProtocolWriter;
import lightcontainer.protocol.rules.reader.SigninRule;
import lightcontainer.protocol.rules.reader.SignupRule;
import lightcontainer.protocol.rules.writer.SignErrorRule;
import lightcontainer.protocol.rules.writer.SignOkRule;

import java.io.*;
import java.net.Socket;
import java.nio.charset.StandardCharsets;

/**
 * ClientHandler
 *
 * <!> UNICAST CLIENT <!>
 * Class communicating with the client, and
 * intercepting and sending files to the client.
 *
 * @version     1.1
 * @since       1.0
 *
 * @see         Runnable
 * @see         AutoCloseable
 * @author      Jérémi NIHART <j.nihart@student.helmo.be>
 */
public class ClientHandler implements Runnable, AutoCloseable {
    // Variables
    private ClientHandlerFFE fileFrontEnd;
    private final Socket client;
    private ProtocolRepository protocolRep;
    private Context context;
    private boolean client_run;


    private BufferedReader reader;
    private PrintWriter writer;
    private ProtocolWriter.ProtocolResult response;

    // Constructor
    public ClientHandler(Socket client, ClientHandlerFFE ffe, ProtocolRepository protocolRep, Context context) {
        this.fileFrontEnd = ffe;
        this.client = client;
        this.protocolRep = protocolRep;
        this.context = context;
        this.client_run = false;
        initClient();
    }

    /**
     * Initialise the Client's Reader and Writer.
     *
     * @since       1.0
     *
     * @see         BufferedReader
     * @see         PrintWriter
     */
    private void initClient() {
        try {
            this.reader = new BufferedReader(new InputStreamReader(
                this.client.getInputStream(),
                StandardCharsets.UTF_8
            ));
            this.writer = new PrintWriter(new OutputStreamWriter(
                this.client.getOutputStream(),
                StandardCharsets.UTF_8
            ), true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    /**
     * Thread Function
     * Start the dialogue with the client.
     *
     * @since       1.0
     */
    @Override
    public void run() {
        this.client_run = true;
        while (this.client_run) {
            try {
                String command = this.reader.readLine();
                if (command != null) {
                    System.out.println("Client: " + command);
                } else this.client.close();

                ProtocolReader.ProtocolResult ruleResult = protocolRep.executeReader(context, command + "\r\n");
                if (ruleResult == null) {
                    this.close();
                    return;
                }

                if (checkAccess(ruleResult)) {
                    ruleResult.read(
                            this.client.getInputStream()
                    );
                    ProtocolWriter.ProtocolResult writerCommand = ruleResult.getResultCommand();

                    if (ruleResult.getReceiver() == ProtocolReader.ResultCmdReceiver.STOREBACKEND) {
                        fileFrontEnd.newCommand(context, writerCommand); // Envoie dans la file de tâche FileFrontEnd en attente d'un traitement d'un StorBackEnd

                        // Attend la fin de la réalisation de la tâche
                        waitTaskResponse();

                        writer.write(response.getCommand()); // Renvoye au client
                        writer.flush();
                        response.write(this.client.getOutputStream()); // Ecrit au client si nécessaire
                    } else {
                        writer.print(writerCommand.getCommand()); // Renvoye au client
                        writer.flush();
                    }

                } else {
                    accessDenied();
                }
            } catch (IOException ignore) {
                ignore.printStackTrace();
            }

        }
    }

    /**
     * Permet de vérifier si le client possède l'accès demandé
     * @param ruleResult La règle
     * @return TRUE si le client possède l'accès demandé
     */
    private boolean checkAccess(ProtocolReader.ProtocolResult ruleResult) {
        if (context.isConnected())
            return true;

        try {
            ruleResult
                    .getClass()
                    .asSubclass(SignupRule.Result.class);
            return true;
        } catch (ClassCastException e) {
            try {
                ruleResult.getClass().asSubclass(SigninRule.Result.class);
                return true;
            } catch (ClassCastException e2) { }
        }

        return false;
    }

    /**
     * Lorsque l'accès à été refusé.
     */
    private void accessDenied() {
        ProtocolWriter.ProtocolResult signErrorResult = protocolRep.executeWriter(SignErrorRule.NAME);
        writer.write(signErrorResult.getCommand()); // Envoie SignError car echec de la connection
        writer.flush();
    }

    /**
     * Permet au Client d'attendre la fin de la réalisation de sa tâche
     */
    private void waitTaskResponse() {
        synchronized (this) {
            try {
                this.wait();
            } catch (InterruptedException e) { e.printStackTrace(); }
        }
    }

    /**
     * Permet d'envoyer la réponse au client.
     * @param response La réponse
     */
    public void respond(ProtocolWriter.ProtocolResult response) {
        this.response = response;
        synchronized (this) {
            this.notify();
        }
    }



    /**
     * AutoClosable Function
     * Close the Client thread and resources.
     *
     * @since       1.0
     */
    @Override
    public void close() {
        if (this.client_run) {
            try {
                this.client_run = false;
                this.client.close();
            } catch (IOException ignored) { }
        }
    }

    public String getLogin() {
        return this.context.getLogin();
    }
}
