package lightcontainer.protocol;

import lightcontainer.domains.client.Context;

import java.io.InputStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public abstract class ProtocolReader {
    private final String name;
    // Variables
    private final Pattern rulePattern;

    // Constructor
    protected ProtocolReader(String name, String pattern) {
        this.name = name;
        this.rulePattern = Pattern.compile(pattern);
    }

    public enum ResultCmdReceiver {
        CLIENT,
        STOREBACKEND
    }

    /**
     * Modèle utilisé par tout les résultats des règles de protocol.
     * Lorsqu'il retourne son résultat, on vérifie si il y a une demande de lecture/écriture de fichier depuis le réseau. Si oui on appel ces méthodes, sinon on ne fait rien.
     * Ensuite on regarde après l'exécution de ces méthode ou non si il y a une commande de retour ou non et l'envoyons au receiver spécifié par la commande.
     */
    public class ProtocolResult {
        /**
         * Command qui sera renvoyée par exemple au client
         */
        private ProtocolWriter.ProtocolResult resultCommand;

        /**
         * Le context courant
         */
        private Context context;

        public ProtocolResult(Context context) {
            this.context = context;
        }

        /**
         * Désigne vers ou cette commande est envoyée.
         * ResultCmdReceiver.CLIENT : Signifie que cette commande va être directement revoyée au client.
         * ResultCmdReceiver.STOREBACKEND : Signifie que cette commande va être envoyée dans la file de tâche du server et êre en attente d'envoie à un StorBackEnd
         */
        private ResultCmdReceiver receiver;

        public ResultCmdReceiver getReceiver() {
            return receiver;
        }

        /**
         * Récupérer la commande à envoyer
         *
         * @return Commande
         */
        public ProtocolWriter.ProtocolResult getResultCommand() {
            return this.resultCommand;
        }

        /**
         * Mettre la commande à envoyer
         *
         * @param resultCommand Commande à envoyer
         * @param receiver      Le receveur de cette commande
         */
        public void setResultCommand(ProtocolWriter.ProtocolResult resultCommand, ResultCmdReceiver receiver) {
            this.resultCommand = resultCommand;
            this.receiver = receiver;
        }

        /**
         * Permet de lire un fichier. Cad reçevoir le contenu d'un fichier provenant du réseau.
         * Redéfinissez cette méthode pour l'utiliser
         *
         * @param reader Buffer rempli du fichier
         */
        public void read(InputStream reader) {
        }

        /**
         * Permet de récupérer le context courant
         * @return Context courant
         */
        protected Context getContext() {
            return context;
        }

    }

    /**
     * Permet de lancer la décomposition d'une commande pour en extraire les données
     *
     * @param data Contenu de la commande
     */
    public <T extends ProtocolResult> T execute(Context context, String data) {
        String name = extractName(data);

        if (name != null && name.equals(this.name)) {
            Matcher ruleMatcher = this.rulePattern.matcher(data);
            // Vérifier que la commande match
            if (ruleMatcher.matches()) {
                String[] groups = new String[ruleMatcher.groupCount()];

                for (int i = 1; i <= groups.length; ++i)
                    groups[i - 1] = ruleMatcher.group(i);

                return onExecuted(context, groups);
            } else {
                return onError(context);
            }
        }
        return null;
    }

    private String extractName(String data) {
        String name;
        int endIndex = data.indexOf(' ');
        if (endIndex > 0) {
            name = data.substring(0, endIndex);
        } else {
            endIndex = data.indexOf('\r');
            name = data.substring(0, endIndex);
        }
        return name;
    }

    /**
     * Cette méthode est appelée lors de l'exécution de la règle
     *
     * @param data Paramètres pour créer la commande.
     */
    protected abstract <T extends ProtocolResult> T onExecuted(Context context, String... data);

    /**
     * Cette méthode est appelée lors d'une erreur de la règle
     */
    protected <T extends ProtocolResult> T onError(Context context) {
        return null;
    }


}
